# compute derivative for bi-level optimization
import numpy as np 
import jax.numpy as jnp
from jax import grad, jit, vmap, jacfwd, jacrev, random, lax
from functools import partial
import jax
from optimizer import adam

# Find g(x) = argmin_y f(x, y) using gradient descent
def find_gx(f, x, y, step_size=0.01, max_iters=10):
    grad_f = grad(f, argnums=1)
    for _ in range(max_iters):
        y -= step_size * grad_f(x, y)
    return y

# Function to compute dg(x)/dx
def dg_dx(f, x, y_init, max_iters=10):
    g_x = find_gx(f, x, y_init, max_iters=max_iters)
    f_xy = jacrev(jacrev(f, argnums=1), argnums=0)(x, g_x)
    f_yy = jacrev(jacrev(f, argnums=1), argnums=1)(x, g_x)
    return -f_xy / f_yy, g_x

# total derivative of f1(x, y) w.r.t. x
def df_dx(f1, f2, x, y, max_iters=10):
    dg, y = dg_dx(f2, x, y, max_iters=max_iters)
    return jacrev(f1, argnums=0)(x, y) + jacrev(f1, argnums=1)(x, y) * dg, y

def find_hx(f1, f2, X, step_size=0.01, max_iters=100):
    x, y, z = X
    for _ in range(max_iters):
        dy, z = df_dx(partial(f1, x), partial(f2, x), y, z)
        y -= step_size * dy
    return y, z
