import numpy as np 
import jax.numpy as jnp
from jax import grad, jit, vmap, jacfwd, jacrev, random, lax
from functools import partial
import jax
from trilevel.implicit_bilevel import df_dx as second_solve
from optimizer import sgd, adam
import matplotlib.pyplot as plt
from tqdm import tqdm
from trilevel.implicit_bilevel import find_hx as suanle

batch_solve = vmap(second_solve, in_axes=(None, None, 0, 0))
suanle_batch_solve = vmap(suanle, in_axes=(None, None, 0))
key = jax.random.PRNGKey(0)  # You can use any integer seed

def f1(x, y, z):
    return -x * (1 - y - z) + x * jnp.exp(x) - x

def f2(x, y, z):
    return -y * (1 - x - z) + y * jnp.exp(y) - y

def f3(x, y, z):
    return -z * (1 - x - y) + z * jnp.exp(z) - z

def find_Z(f, x, y, z, step_size=0.01, max_iters=100):
    def sum_f(x, y, z):
        return jnp.sum(f(x, y, z))
    opt = None
    grad_f = grad(sum_f, argnums=2)
    key = jax.random.PRNGKey(2024) 
    key, subkey = random.split(key)
    for _ in range(max_iters):
        z, opt = adam(opt, z, grad_f(x, y, z), subkey)
    return z

def solve(X, max_iters=100):
    x, y, z = X[:, 0], X[:, 1], X[:, 2]
    key = jax.random.PRNGKey(2024)  # You can use any integer seed
    opt = None
    for i in tqdm(range(max_iters)):
        dy, z = batch_solve(partial(f2, x), partial(f3, x), y, z)
        key, subkey = random.split(key)
        y, opt = adam(opt, y, dy, subkey)
    return y, z

Data = np.load("Loss.npz") # 500 * 100
select, drop= 5, 50
ITD = Data["ITD"][drop::select]
AID = Data["AID"][drop::select]
VGD = Data["VGD"][drop::select]
data = ITD.reshape(-1)
data2 = jnp.stack([data, jnp.zeros_like(data, dtype=float), jnp.zeros_like(data, dtype=float)], axis=1)
# y, z = solve(data2)
y, z = suanle_batch_solve(f2, f3, data2)
z = find_Z(f3, data, y, z)
y, z = y.reshape((500-drop) // select, 100), z.reshape((500-drop) // select, 100)
loss = jnp.mean(f1(ITD, y, z), axis=1)
plt.plot(np.arange(drop, 500, select), loss, label="Iterative Differentiation")

data = AID.reshape(-1)
data2 = jnp.stack([data, jnp.zeros_like(data, dtype=float), jnp.zeros_like(data, dtype=float)], axis=1)
# y, z = solve(data2)
y, z = suanle_batch_solve(f2, f3, data2)
z = find_Z(f3, data, y, z)
y, z = y.reshape((500-drop) // select, 100), z.reshape((500-drop) // select, 100)
loss = jnp.mean(f1(AID, y, z), axis=1)
plt.plot(np.arange(drop, 500, select), loss, label="Implicit Differentiation")

data = VGD.reshape(-1)
data2 = jnp.stack([data, jnp.zeros_like(data, dtype=float), jnp.zeros_like(data, dtype=float)], axis=1)
# y, z = solve(data2)
y, z = suanle_batch_solve(f2, f3, data2)
z = find_Z(f3, data, y, z)
y, z = y.reshape((500-drop) // select, 100), z.reshape((500-drop) // select, 100)
loss = jnp.mean(f1(VGD, y, z), axis=1)
plt.plot(np.arange(drop, 500, select), loss, label="Vanilla Gradient Descent")

plt.xlabel('Optimization Steps')
plt.ylabel('$f_1(x, h(x), g(x,h(x)))$')
plt.tight_layout()
plt.legend()
plt.savefig('Opti.jpg')