import numpy as np 
import jax.numpy as jnp
from jax import grad, jit, vmap, jacfwd, jacrev, random, lax
from functools import partial
import jax
from trilevel.implicit import wrap_df as implicit_solve
from vanilla import wrap_df as vanilla_solve
from optimizer import sgd, adam
from trilevel.iterative import g_1 as iterative_solve
import matplotlib.pyplot as plt
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('--probllem', type=str, default="first")
parser.add_argument('--sample_size', type=int, default=100)
parser.add_argument('--max_iters', type=int, default=500)
args = parser.parse_args()


third_batch_solve = vmap(implicit_solve, in_axes=(None, None, None, 0, 0, 0))
vanilla_batch_solve = vmap(vanilla_solve, in_axes=(None, None, None, 0, 0, 0))
sota_batch_solve = vmap(iterative_solve, in_axes=(None, None, None, 0, 0, 0))
key = jax.random.PRNGKey(0)  # You can use any integer seed
key, subkey = random.split(key)
X = jax.random.uniform(subkey, shape=(args.sample_size, 3))

if args.probllem == "first":
    def f1(x, y, z):
        return (x-z)**2 + x ** 2

    def f2(x, y, z):
        return (y-x) **2

    def f3(x, y, z):
        return (z-y) **2
    
elif args.probllem == "second":
    def f1(x, y, z):
        return -x * (1 - x - y - z)

    def f2(x, y, z):
        return -y * (1 - x - y - z)

    def f3(x, y, z):
        return -z * (1 - x - y - z)

elif args.probllem == "third":
    def f1(x, y, z):
        return -x * (1 - y - z) + x * jnp.exp(x) - x

    def f2(x, y, z):
        return -y * (1 - x - z) + y * jnp.exp(y) - y

    def f3(x, y, z):
        return -z * (1 - x - y) + z * jnp.exp(z) - z


def solve(batch_solve, X, max_iters=args.max_iters):
    x, y, z = X[:, 0], X[:, 1], X[:, 2]
    key = jax.random.PRNGKey(2024)  # You can use any integer seed
    opt = None
    Loss = []
    for i in range(max_iters):
        dx, y, z = batch_solve(f1, f2, f3, x, y, z)
        key, subkey = random.split(key)
        x, opt = adam(opt, x, dx, subkey)
        if (i+1) % 1 == 0:
            Loss.append(x)
    return x, y, z, Loss

import time
s = time.time()
x, y, z, Loss = solve(sota_batch_solve, X)
print(time.time() - s)
plt.plot(Loss, label="Iterative Differentiation")
ITD = Loss

s = time.time()
x, y, z, Loss = solve(third_batch_solve, X)
print(time.time() - s)
plt.plot(Loss, label="Implicit Differentiation")
AID = Loss

s = time.time()
x,y,z, Loss = solve(vanilla_batch_solve, X)
print(time.time() - s)
plt.plot(Loss, label="Vanilla Gradient Descent")
VGD = Loss

plt.xlabel('Optimization Steps')
plt.ylabel('MSE Loss')
plt.legend()
plt.savefig('Loss.jpg')

# np.savez("Loss.npz", ITD=ITD, AID=AID, VGD=VGD)