import numpy as np 
import jax.numpy as jnp
from jax import grad, jit, vmap, jacfwd, jacrev, random, lax
from functools import partial
import jax

def sgd(opt, x, dx, key):
    if opt == None:
        opt = dict()
        opt['step_size'] = 0.01
        opt['noise_std'] = 0.01
    return x - opt['step_size'] * (dx + jax.random.normal(key, shape=dx.shape) * opt['noise_std']), opt

def adam(opt, x, dx, key=None):
    if opt == None:
        opt = dict()
        opt['m'] = 0
        opt['v'] = 0
        opt['t'] = 0
        opt['beta1'] = 0.9
        opt['beta2'] = 0.999
        opt['eps'] = 1e-8
        opt['step_size'] = 0.01
        opt['noise_std'] = 0.01
    m, v, t = opt['m'], opt['v'], opt['t']
    beta1, beta2, eps = opt['beta1'], opt['beta2'], opt['eps']
    m = beta1 * m + (1 - beta1) * dx
    v = beta2 * v + (1 - beta2) * dx**2
    opt['m'], opt['v'], opt['t'] = m, v, t+1
    m_hat = m / (1 - beta1**(t+1))
    v_hat = v / (1 - beta2**(t+1))
    return x - opt['step_size'] * (m_hat / (jnp.sqrt(v_hat) + eps) + jax.random.normal(key, shape=dx.shape) * opt['noise_std']), opt