import numpy as np 
import jax.numpy as jnp
from jax import grad, jit, vmap, jacfwd, jacrev, random, lax
from functools import partial
import jax

# Find g(x) = argmin_y f(x, y) using gradient descent
def find_gx(f, x, y, step_size=0.01, max_iters=10):
    grad_f = grad(f, argnums=1)
    for _ in range(max_iters):
        y -= step_size * grad_f(x, y)
    return y

# total derivative of f1(x, y) w.r.t. x
def df_dx(f1, f2, x, y):
    y = find_gx(f2, x, y, max_iters=20)
    return jacrev(f1, argnums=0)(x, y), y

def find_hx(f1, f2, X, step_size=0.01, max_iters=1000):
    x, y, z = X
    for _ in range(max_iters):
        dy, z = df_dx(partial(f1, x), partial(f2, x), y, z)
        y -= step_size * dy
    return y, z

def wrap_df(f1, f2, f3, x, y, z):
    y, z = find_hx(f2, f3, jnp.stack([x,y,z], axis=0), max_iters=10)
    z = find_gx(partial(f3, x), y, z, max_iters=10)
    dx = jacrev(f1, argnums=0)(x, y, z)
    return dx, y, z